/*******************************************************************
made by Johannes Philippi, Feb. 2007
johannesphilippi[a t]web[d o t]de

This is a very simple program which demonstrates basic 
CAN functionality:
	- configuring the CAN driver (250kBit/s)
	- configuring MessageObjects (in extended frame format)
	- sending a message on the CAN (switch-controlled)
	- receiving a message from the CAN (interrupt-controlled)
	- handling ID-tags and masks
The display is ued to output some information (incoming IDs) and 
for signaling transmit and receive activities.
*******************************************************************/

#include <avr/io.h>           	//names of bytes and bits of I/O registers
#include <avr/signal.h>			//Interrupt vectors of uC
#include <inttypes.h>			//AVR style type definitions
#include <avr/interrupt.h>
#include "delaymsCAN128.c"
#include "displayDef.h"


/*******************************************************************
definitions
*******************************************************************/
#define NO_MOBS 15		// number of message objects
#define NOMOB   0xff

//definitions for AMIS dual CAN transceiver at PORTD (Enable1, Enable2)
#define ENB1 PD7
#define ENB2 PD4

#define setbit(ADR,BIT) (ADR|=(1<<BIT))
#define clearbit(ADR,BIT) (ADR&=~(1<<BIT))

// CAN2.0 B, extendef frame(29bit) message
typedef struct
	{	
		uint32_t id;			// message ID                  
		uint32_t msk;        	// ID Mask for message          
		uint8_t length;   		// length of data DLC
		uint8_t data[8];  		// data vector
	}CAN_messageType;


/*******************************************************************
globals
*******************************************************************/

CAN_messageType recMsg;


/*******************************************************************
Get idex of MOb for which an interrupt was detected.
*******************************************************************/
uint8_t getmob( uint32_t bitmask) 
{
 uint8_t mob;

 if( bitmask==0) return NOMOB;

 //the bitmask is shifted to the rigth until the LSB is one
 //the number of such shift operations is the index of the MOb
 for( mob=0; (bitmask & 0x01)==0; bitmask >>= 1, ++mob);

 //security check
 if ( mob > 14 ) return NOMOB;
 	else return mob;
}     

/*******************************************************************
CAN Receive Interrupt Service Routine
*******************************************************************/
SIGNAL(SIG_CAN_INTERRUPT1)
{
//CAN_messageType recMsg; this is a global variable
uint8_t save_canpage = CANPAGE;		// store current CANPAGE
uint8_t mob,i;

	// get the index of the MOb for which the interrupt occured
	mob = getmob(CANSIT2 | (CANSIT1 << 8));
		if( mob == NOMOB )   return;
   	
	//select concerned MOb page
	CANPAGE = mob << 4;	

	setbit(PORTC,PC0); 				//set sign for being arrived at this point

/*	// debug-code
	lcd_write(p11,0);
	lcd_write_number(CANSTMOB);
	lcd_write_string("#");
	lcd_write_number(CANGIE);
	lcd_write_string("#");
	lcd_write_number(CANGIT);
	lcd_write(p21,0);
	lcd_write_number(CANTEC);
	lcd_write_string("#");
	lcd_write_number(CANSIT2);
	lcd_write_string("#");
*/
	if (CANSTMOB & (1<<TXOK))
		{
		 setbit(PORTC,PC1);			// signaling txOK
	 	 lcd_write(p115,0);
	 	 lcd_write_string("tx");
		 delayms(150);
		 lcd_write(p115,0);
	 	 lcd_write_string("  ");
		 	
		 CANSTMOB &= ~(1<<TXOK);	// reset INT reason
				
		}
	else if (CANSTMOB & (1<<RXOK))
		{
	 	 setbit(PORTC,PC2);			// signaling rxOK
		 lcd_write(p115,0);
	 	 lcd_write_string("rx");
		 delayms(150);
		 lcd_write(p115,0);
	 	 lcd_write_string("  ");

			 //get data of selected MOb
			 for (i=0; i<8; i++)		
		 		 recMsg.data[i] = CANMSG;
	
			 //get ID of selected MOb
			 recMsg.id = 0;
			 recMsg.id |= ((uint32_t) CANIDT1<<24);	//get identifier
			 recMsg.id |= ((uint32_t) CANIDT2<<16);
			 recMsg.id |= ((uint32_t) CANIDT3<<8);
			 recMsg.id |= (CANIDT4&0xF8);
			 recMsg.id >>= 3;

		 	 lcd_write(p19,0);						//display low byte of ID of recMsg
	 	 	 lcd_write_string("ID:");
	 	 	 lcd_write_number((uint8_t)recMsg.id);	//display low byte of ID of recMsg

		 CANSTMOB &= ~(1<<RXOK);					// reset INT reason 	
		 CANCDMOB = (1<<CONMOB1) | (1<<IDE) | (8); 	// re-enable RX, reconfigure MOb IDE=1
		}
	else 
		{
		 setbit(PORTC,PC3);
		 delayms(150);		
		}
		
 
/* 	//debug-code
 	lcd_write(p29,0);
	lcd_write_number(CANSTMOB);
*/
    PORTC = 0x00;				// clear all LEDs 
	
	CANPAGE = save_canpage;   	// restore CANPAGE

}


/*******************************************************************
CAN initialization
*******************************************************************/
void can_init( void)
{
uint8_t mob;

  CANGCON |= (1<<SWRES);		// reset the CAN controller
  
  CANGCON = 0x00;				// reset the general control register
  CANGIT  = 0x00;				// reset the general interrupt register
   	
  CANSIT2 = 0x00;				// reset the status interrupt registers
  CANSIT1 = 0x00;

  //Enable CAN interrupts globally, set enable all interrupts except CANtimer overrun interrupt
  CANGIE = 0xfe;	//(1<<ENIT) | (1<<ENRX) | (1<<ENTX) |(1<<ENERR);//
  
  //Set MObs for which interrupts are enabled
  CANIE2 = (1<<IEMOB0) |  (1<<IEMOB1) |  (1<<IEMOB2) |  (1<<IEMOB3); 		
  CANIE1 = 0;

  //for all Message Objects: set the StatusRegister CANSTMOB to zero; Disable MOb CANCDMOB=0, (CONMOB1:0=00)
  for (mob = 0; mob < NO_MOBS; mob++)
  {
     CANPAGE  = (mob << 4);
	 CANIDT1 = 0x00;		//reset ID-Tag
	 CANIDT2 = 0x00;
	 CANIDT3 = 0x00;
	 CANIDT4 = 0x00;

	 CANIDM1 = 0x00;		//reset ID-Mask
	 CANIDM2 = 0x00;
	 CANIDM3 = 0x00;
	 CANIDM4 = 0x00;

     CANSTMOB = 0x00;		//reset MOb status
     CANCDMOB = 0x00; 		//disable MOb
  }

	
	
  CANBT1 = 0x06;	// set Baud Rate Prescaler for CAN controller bit timing (page236, table on page261)
  CANBT2 = 0x0C;	// set propagation time segment (page232, table on page261)
  CANBT3 = 0x37;	// set PHaseSegment2 and PHaseSegment1 (page233, table on page261)
   
  CANTCON = 0xFF; 	// CAN timer prescaler 256. Can be used for time stamping
       
  CANGCON |= (1<<ENASTB);	// Set CAN-driver in enable mode (not standby-mode)
  	
  // wait until module is ready EnableFlag is on.
  while (!(CANGSTA & (1<<ENFG))); 

//---------------- CAN CONTROLLER IS NOW INITIALISED/CONFIGURED 250kBits/s ---------------
  
  // this part must be adapted to your hardware!!
  // set enable CAN-channel 1 of amis42700 dual CAN-transceiver 
  DDRD |= (1<<ENB1) | (1<<ENB2);
  PORTD &= ~(1<<ENB1); 			// enable channel1.  (low-active type)
  PORTD |= (1<<ENB2);			// disable channel2. (low-active type)
    
}//can_init


/*******************************************************************
CAN transmission 
*******************************************************************/
void can_tx (CAN_messageType msg)
{	
uint8_t i, mob;	

	//select MOb to be sent
	mob=3;

	//enable MOb number mob, auto increment index, start with index = 0
	CANPAGE = (mob<<4);
	
	//set IDE bit, length = 8
	CANCDMOB = (1<<IDE) | (8);			
	
	//write 29 Bit identifier to MOb-ID-Tag
	msg.id <<= 3;								
	CANIDT4 = (unsigned char) (msg.id&0xF8);
	CANIDT3 = (unsigned char) (msg.id>>8);
	CANIDT2 = (unsigned char) (msg.id>>16);
	CANIDT1 = (unsigned char) (msg.id>>24);
	
	//put data in mailbox
	for (i=0; i<8; i++)							
		CANMSG = msg.data [i];
	
	//wait until transmit possible (TXBSY==0)
	while (CANGSTA & (1<<TXBSY));
	
	//enable transmission. After the tx is complete the TXOK bit is cleared in the ISR
	CANCDMOB |= (1<<CONMOB0);											

}//can_tx


/*******************************************************************
MAIN PROGRAM ENTRY POINT
*******************************************************************/
int main (void) 
{
CAN_messageType msg,msg0;
uint8_t aux=1,mob;

	//init for switch Portpin. (active-low type) 
	DDRF &= !(1<<PF2);	//(set input pin) see datasheet p62/63
	PORTF |= (1<<PF2);	//(activate pull-up)
	
	DDRC = 0xff;					//PC0...PC7 als Ausgnge definieren fr LEDs
	PORTC= 0x00;

	lcd_init();						//in lcd_init wird auch der Timer fr delayms(xx) initialisiert.
		
	lcd_clear();
	lcd_write(p11,0);
	lcd_write_string("CANtest");
	lcd_write(p21,0);
	lcd_write_string("J.Philippi 02/07");
	
	can_init();						//initialize CAN driver
	sei();							//globally enable interrupts

// create CAN_package to be sent
	msg.id = 0x80;
	msg.length = 8;
	msg.data[0] = 0x01;
	msg.data[1] = 0x02;
	msg.data[2] = 0x01;
	msg.data[3] = 0x02;
	msg.data[4] = 0x01;
	msg.data[5] = 0x02;
	msg.data[6] = 0x01;
	msg.data[7] = 0x02;
	
// select MOb to be configured for receiving a message with a certain ID messages --------
	mob = 1;

	//configure  MOb
	CANPAGE = (mob<<4);				// select MOb page	
	
	msg0.id = 0x00000091;			// define an ID for the message
	CANIDT4 = msg0.id << 3;		
	CANIDT3 = msg0.id >> 5;	
	CANIDT2 = msg0.id >> 13;	
	CANIDT1 = msg0.id >> 21;	

	// define an ID mask for the MOb. 
	msg0.msk = 0xFFFFFFFF;			// set IDmask to receive (only ID 0x91)
	CANIDM4 = (msg0.msk << 3) | (1<<IDEMSK); // set IDEMSK to 1 -> extended identifier
	CANIDM3 = msg0.msk >> 5;
	CANIDM2 = msg0.msk >> 13;
	CANIDM1 = msg0.msk >> 21;

	// configure message as receive-msg, Extended frame format, data length code=8 
	// (see CANCDMOB register, page257)
	CANCDMOB = (1<<CONMOB1) | (1<<IDE) | (0x08);
// --------------------------------------------------------------------------------------------
	
//select MOb to be configured for receiving messages with some more IDs -----------------------
	mob = 2;

	//configure  MOb 
	CANPAGE = (mob<<4);				// select MOb page	
	
	msg0.id = 0x80;			// define an ID for the message
	CANIDT4 = (msg0.id << 3) ;		
	CANIDT3 = msg0.id >> 5;	
	CANIDT2 = msg0.id >> 13;	
	CANIDT1 = msg0.id >> 21;	

	//receive all messages on the bus
	msg0.msk = 0xFFFFFFF0;		// set IDmask to receive IDs from 0x80...0x8F
	CANIDM4 = (msg0.msk << 3)| (1<<IDEMSK);	// set IDEMSK to 1 -> extended identifier
	CANIDM3 = msg0.msk >> 5;
	CANIDM2 = msg0.msk >> 13;
	CANIDM1 = msg0.msk >> 21;

	// configure message as receive-msg, Extended frame format, data length code=8 
	// (see CANCDMOB register, page257)
	CANCDMOB = (1<<CONMOB1) | (1<<IDE) | (0x08);
// --------------------------------------------------------------------------------------------


	//mark initialisation finished. Blink LED.
	PORTC |= (1<<PC2);
	delayms(500);
	PORTC &= ~(1<<PC2);

	
	// inf. program loop
	for (;;) 						
	{
		
		// this part has to eb adapted to your HW.
		// check switch state. If closed send message once only
		if (!(PINF & (1<<PINF2)) )			
			{
				lcd_write(p116,0);	
				lcd_write_string("*");
				if (aux ==1) 
					{
					 can_tx(msg);
					 aux=0;
					}
			}
			else 
			{
				lcd_write(p116,0);	
				lcd_write_string(" ");
				aux=1;
			}
		
	} //for
}//main
